package corrodias.sr3.save.data;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class ListReader {

	private List<String> weaponNames;
	private List<Territory> territories;

	public ListReader() throws Exception {
		weaponNames = loadWeaponNameList();
		territories = loadTerritoryList();
		StringHashes.loadUpgradeList();
		StringHashes.loadStringList("weapon_list.txt");
		StringHashes.loadStringList("homie_list.txt");
	}

	private List<String> loadWeaponNameList() throws Exception {
		BufferedReader weaponStream = null;
		try {
			try {
				weaponStream = new BufferedReader(new InputStreamReader(new FileInputStream("weapon_list.txt"), StandardCharsets.UTF_8));
			} catch (FileNotFoundException ex) {
				weaponStream = new BufferedReader(new InputStreamReader(ClassLoader.getSystemResourceAsStream("corrodias/sr3/save/data/lists/weapon_list.txt"), StandardCharsets.UTF_8));
			}
			List<String> result = new ArrayList<>(100);
			String line;
			while ((line = weaponStream.readLine()) != null) {
				result.add(line);
			}
			return result;
		} finally {
			try {
				weaponStream.close();
			} catch (Exception ex2) {
			}
		}
	}

	private List<Territory> loadTerritoryList() throws Exception {
		BufferedReader territoryStream = null;
		try {
			try {
				territoryStream = new BufferedReader(new InputStreamReader(new FileInputStream("territory_list.txt"), StandardCharsets.UTF_8));
			} catch (FileNotFoundException ex) {
				territoryStream = new BufferedReader(new InputStreamReader(ClassLoader.getSystemResourceAsStream("corrodias/sr3/save/data/lists/territory_list.txt"), StandardCharsets.UTF_8));
			}
			List<Territory> result = new ArrayList<>(200);
			String line;
			while ((line = territoryStream.readLine()) != null) {
				String[] linePieces = line.split(":");

				Territory territory = new Territory();
				territory.district = District.valueOf(linePieces[0].replace(" ", "_"));
				territory.neighborhood = Neighborhood.valueOf(linePieces[1].replace(" ", "_"));
				territory.name = linePieces[2];

				result.add(territory);
			}
			return result;
		} finally {
			try {
				territoryStream.close();
			} catch (Exception ex2) {
			}
		}
	}

	public List<String> getWeaponNames() {
		return Collections.unmodifiableList(weaponNames);
	}

	public List<Territory> getTerritories() {
		return Collections.unmodifiableList(territories);
	}

	public void showGarageEditor() throws Exception {
	}
}
